/* MEX function to solve SDP 

*/
// Last Modified: A. d'Aspremont, Feb 2009
// http://www.carva.org/alexandre.daspremont


#include "ColSampGrad.h" 

void mexFunction(int nlhs, mxArray *plhs[], int nrhs, const mxArray *prhs[])
{
	// For main loop
	double *Amat,*bvec,*cvec,*xvec,*avxvec,*grad,*ovals,*movals,*gaps,Bnd,ps,stepscal;
	double *bufveca,*bufmat,*probv,*cprobv,*bufccmat,*bufcovmat,alpha,beta,stepsz,bufnrm,buf,*cput,targetg;
	double *rndn;
	int n,m,i,j,k,l,n2,cc,idum=-1,*sidx;
	int info,Nperiod,maxiter;
	double cputime;int left_h=0,left_m=0,left_s=0;
	clock_t start_time=(double)clock();
	// For LAPACK
	int work_size; 
	int lwork,inflapack,indmax;
	double *workvec;
	char jobz[1]="V",uplo[1]="U";
	// For Arpack & BLAS
	double *evec,eigv=0;
	int nummatvec,nconv,arcount,ncv,nev,incx=1,swsgn;
	char which[2]="LM";double tol=1e-8; int maxitrarp=1000;
		
	// Inputs
	Amat=mxGetPr(prhs[0]);
	m = (int)mxGetN(prhs[0]);
	n2 = (int)mxGetM(prhs[0]);
	bvec = mxGetPr(prhs[1]);
	cvec = mxGetPr(prhs[2]);
	Bnd =(int)mxGetScalar(prhs[3]);
	ps = mxGetScalar(prhs[4]);
	stepscal = mxGetScalar(prhs[5]);
	maxiter =(int)mxGetScalar(prhs[6]);
	info =(int)mxGetScalar(prhs[7]);
	targetg =mxGetScalar(prhs[8]);
	n=(int)sqrt(n2);cc=(int)(ps*n);
	work_size=3*n+n*n;lwork=work_size; // For LAPACK
	// Outputs
	plhs[0] = mxCreateDoubleMatrix(m, 1, mxREAL);
	avxvec = mxGetPr(plhs[0]);
	plhs[1] = mxCreateDoubleMatrix(maxiter+1, 1, mxREAL);
	ovals = mxGetPr(plhs[1]);
	plhs[2] = mxCreateDoubleMatrix(maxiter+1, 1, mxREAL);
	movals = mxGetPr(plhs[2]);
	plhs[3] = mxCreateDoubleMatrix(maxiter+1, 1, mxREAL);
	gaps = mxGetPr(plhs[3]);
	plhs[4] = mxCreateDoubleMatrix(maxiter+1, 1, mxREAL);
	cput = mxGetPr(plhs[4]);


	// Allocate
	evec=(double*)calloc(n,sizeof(double));
	probv=(double*)calloc(n,sizeof(double));
	cprobv=(double*)calloc(n+1,sizeof(double));
	bufveca=(double*)calloc(n,sizeof(double));
	xvec=(double*)calloc(m,sizeof(double));	
	grad=(double*)calloc(m,sizeof(double));	
	bufmat=(double*)calloc(n2,sizeof(double));	
	bufccmat=(double*)calloc(n*cc,sizeof(double));
	bufcovmat=(double*)calloc(cc*cc,sizeof(double));
	rndn=(double*)calloc(cc,sizeof(double));
	sidx=(int*)calloc(cc,sizeof(int));
	#ifndef usearpack
	workvec=(double *) calloc(work_size,sizeof(double)); // For LAPACK
	#endif
	
	// Start main loop
	stepsz=stepscal/sqrt(maxiter);Nperiod=maxiter/info;
	alpha=0.0;cblas_dscal(m,alpha,xvec,incx);
	for(k=0;k<=maxiter;k++){
		// Form current matrix
		alpha=0.0;cblas_dscal(n2,alpha,bufmat,incx);
		for(j=0;j<m;j++){alpha=xvec[j];cblas_daxpy(n2,alpha,Amat+j*n2,incx,bufmat,incx);}
		alpha=-1.0;cblas_daxpy(n2,alpha,cvec,incx,bufmat,incx);
		// Compute sampling probability
		bufnrm=doubnorm2(bufmat,n2);bufnrm=bufnrm*bufnrm;
		for(i=0;i<n;i++){probv[i]=doubnorm2(bufmat+i*n,n);probv[i]=probv[i]*probv[i]/bufnrm;}
		cprobv[0]=0.0;cprobv[1]=probv[0];for(i=1;i<n;i++){cprobv[i+1]=cprobv[i]+probv[i];}
		// Sample from it
		for (i=0;i<cc;i++){rndn[i]=ran1(&idum);}
		for(i=0;i<cc;i++){sidx[i]=n-1;j=n-2;while(rndn[i]<cprobv[j+1]){sidx[i]=j;j--;}}
		// Create subsampled matrix
		for(i=0;i<cc;i++){
			cblas_dcopy(n,bufmat+sidx[i]*n,incx,bufccmat+i*n,incx);
			alpha=1.0/sqrt(cc*probv[sidx[i]]);cblas_dscal(n,alpha,bufccmat+i*n,incx);}
		// Compute eg decomposition
		#ifdef usearpack // Switch between ARPACK and LAPACK
		nev=1;ncv=4; // TODO: Should get a better size estimate here (cf. arpack manual)
		info=simlrarpack(bufccmat,n,cc,nev,ncv,tol,which,maxitrarp,0,&eigv,bufveca,&nconv,&nummatvec,&arcount);
		#else
		// Square it 
		alpha=1.0;beta=0.0;
		cblas_dgemm(CblasColMajor,CblasTrans,CblasNoTrans,cc,cc,n,alpha,bufccmat,n,bufccmat,n,beta,bufcovmat,cc);
		dsyev(jobz,uplo,&cc,bufcovmat,&cc,bufveca,workvec,&lwork,&inflapack);
		indmax=idxabsmax(bufveca,cc);eigv=bufveca[indmax];cblas_dcopy(cc,bufcovmat+indmax*cc,incx,bufveca,incx);
		#endif
		// Reconstruct eigenvector
		alpha=1.0/sqrt(eigv);beta=0.0;
		cblas_dgemv(CblasColMajor,CblasNoTrans,n,cc,alpha,bufccmat,n,bufveca,incx,beta,evec,incx);
		// Get gradient and step
		for(j=0;j<m;j++){
			buf=0;
			for(i=0;i<cc;i++){for(l=0;l<n;l++){buf+=Amat[j*n2+sidx[i]*n+l]*(evec[sidx[i]]*evec[l])/(cc*probv[sidx[i]]);}}
			grad[j]=buf;}
		// Estimate sign of eigenvalue and finish gradient
		buf=0;for(i=0;i<cc;i++){for(l=0;l<n;l++){buf+=cvec[sidx[i]*n+l]*(evec[sidx[i]]*evec[l])/(cc*probv[sidx[i]]);}}
		swsgn=dsignf(doubdot(grad,xvec,m)-buf);cblas_dscal(m,swsgn,grad,incx);
		alpha=-1.0;cblas_daxpy(m,alpha,bvec,incx,grad,incx);
		// Take step, project & record objective value + avg.
		alpha=-stepsz;cblas_daxpy(m,alpha,grad,incx,xvec,incx);
		buf=doubnorm2(xvec,m);if(buf>1e-12){alpha=dminif(Bnd,buf)/buf;cblas_dscal(m,alpha,xvec,incx);}
		ovals[k]=sqrt(eigv)-doubdot(bvec,xvec,m);
		alpha=k;cblas_dscal(m,alpha,avxvec,incx);
		alpha=1.0;cblas_daxpy(m,alpha,xvec,incx,avxvec,incx);
		alpha=1.0/(k+1.0);cblas_dscal(m,alpha,avxvec,incx);
		// Report progress periodically  *** TODO: subsample this too  ***
		cputime=(double)((clock()-start_time)/CLOCKS_PER_SEC);
		cput[k]=cputime;
		if (k%Nperiod==0){
			alpha=0.0;cblas_dscal(n2,alpha,bufmat,incx);
			for(j=0;j<m;j++){alpha=avxvec[j];cblas_daxpy(n2,alpha,Amat+j*n2,incx,bufmat,incx);}
			alpha=-1.0;cblas_daxpy(n2,alpha,cvec,incx,bufmat,incx);
			#ifdef usearpack // Switch between ARPACK and LAPACK
			nev=1;ncv=4; // TODO: Should get a better size estimate here (cf. arpack manual)
			info=simarpack(bufmat,n,nev,ncv,tol,which,maxitrarp,0,&eigv,evec,&nconv,&nummatvec,&arcount);
			#else
			dsyev(jobz,uplo,&n,bufmat,&n,bufveca,workvec,&lwork,&inflapack); 
			indmax=idxabsmax(bufveca,n);eigv=bufveca[indmax];
			#endif
			movals[k]=dabsf(eigv)-doubdot(bvec,avxvec,m);
			// Compute gap
			for(i=0;i<n;i++){for(j=0;j<n;j++){bufmat[j*n+i]=evec[i]*evec[j];}}
			for(j=0;j<m;j++){grad[j]=doubdot(Amat+j*n2,bufmat,n2);}
			swsgn=dsignf(doubdot(xvec,grad,m)-doubdot(bufmat,cvec,n2));cblas_dscal(m,swsgn,grad,incx);
			alpha=-1.0;cblas_daxpy(m,alpha,bvec,incx,grad,incx);			
			buf=0;for(j=0;j<n;j++){for(i=0;i<n;i++){buf+=cvec[j*n+i]*evec[i]*evec[j];}} // Could also recycle approx. above
			gaps[k]=movals[k]+swsgn*buf+Bnd*doubnorm2(grad,m);
			left_h=(int)floor(cputime/3600);left_m=(int)floor(cputime/60-left_h*60);left_s=(int)floor(cputime-left_h*3600-left_m*60);
			mexPrintf("Iter: %.3e   Obj: %.4e    Gap: %.4e   CPU Time: %2dh %2dm %2ds\n",(double)(k),movals[k],gaps[k],left_h,left_m,left_s);
			mexEvalString("drawnow;");
			//if (gaps[k]/gaps[0]<=targetg){break;}
			if (movals[k]<=targetg){break;}
		}}
	// Free everything
	free(evec);free(bufveca);free(bufmat);free(xvec);free(grad);free(probv);free(cprobv);free(bufccmat);free(bufcovmat);
	free(sidx);free(rndn);
	#ifndef usearpack
	free(workvec);
	#endif
}