/* MEX function to solve SDP 

*/
// Last Modified: A. d'Aspremont, Feb 2009
// http://www.carva.org/alexandre.daspremont


#include "DetDSPCA.h" 

void mexFunction(int nlhs, mxArray *plhs[], int nrhs, const mxArray *prhs[])
{
	// For main loop
	double *Amat;
	double *Xmat,*Xavmat,*grad,*ovals,*movals,*gaps,Bnd,stepscal,*cput,targetg;
	double *probv,*bufmata,eigv,swsgn;
	double alpha,beta,stepsz;
	int n,i,k,n2;
	int info,arpinfo,Nperiod,maxiter;
	double cputime;int left_h=0,left_m=0,left_s=0;
	clock_t start_time=(double)clock();
	double *evec;
	int nummatvec,nconv,arcount,ncv,nev,incx=1;
	char which[2]="LM";double tol=1e-6; int maxitrarp=1000; // ARPACK parameters
		
	// Inputs
	Amat=mxGetPr(prhs[0]);
	n = (int)mxGetN(prhs[0]);
	Bnd =mxGetScalar(prhs[1]); // Bound on ratings
	stepscal = mxGetScalar(prhs[2]);
	maxiter =(int)mxGetScalar(prhs[3]);
	info =(int)mxGetScalar(prhs[4]);
	targetg =mxGetScalar(prhs[5]);
	n2=n*n;
	// Outputs
	plhs[0] = mxCreateDoubleMatrix(n,n, mxREAL);
	Xmat = mxGetPr(plhs[0]);
	plhs[1] = mxCreateDoubleMatrix(n,n, mxREAL);
	Xavmat = mxGetPr(plhs[1]);
	plhs[2] = mxCreateDoubleMatrix(maxiter+1, 1, mxREAL);
	ovals = mxGetPr(plhs[2]);
	plhs[3] = mxCreateDoubleMatrix(maxiter+1, 1, mxREAL);
	movals = mxGetPr(plhs[3]);
	plhs[4] = mxCreateDoubleMatrix(maxiter+1, 1, mxREAL);
	gaps = mxGetPr(plhs[4]);
	plhs[5] = mxCreateDoubleMatrix(maxiter+1, 1, mxREAL);
	cput = mxGetPr(plhs[5]);


	// Allocate
	evec=(double*)calloc(n,sizeof(double));
	probv=(double*)calloc(n,sizeof(double));
	bufmata=(double*)calloc(n,sizeof(double));
	grad=(double*)calloc(n*n,sizeof(double));	

	// Initialize X matrix
	cblas_dcopy(n2,Amat,incx,Xmat,incx);
	// Start main loop
	stepsz=stepscal/sqrt(maxiter);Nperiod=maxiter/info;
	for(k=0;k<=maxiter;k++){
		// Compute eig decomposition
		nev=1;ncv=iminf(3,n); // TODO: Should get a better size estimate for ncv here (cf. arapck manual)
		arpinfo=simlrarpack(Xmat,n,n,nev,ncv,tol,which,maxitrarp,0,&eigv,evec,&nconv,&nummatvec,&arcount);
		// Get sign of eigenvalues
		alpha=1.0;beta=0.0;
		cblas_dgemv(CblasColMajor,CblasNoTrans,n,n,alpha,Xmat,n,evec,incx,beta,probv,incx);
		swsgn=dsignf(doubdot(probv,evec,n));
		if (eigv<=tol){swsgn=0.0;}
		// Form gradient matrix
		cblas_dcopy(n,evec,incx,bufmata,incx);
		alpha=swsgn;cblas_dscal(n,alpha,bufmata,incx);
		alpha=1.0;beta=0.0;
		cblas_dgemm(CblasColMajor,CblasNoTrans,CblasTrans,n,n,1,alpha,evec,n,bufmata,n,beta,grad,n);
		// Take step 
		alpha=-stepsz;cblas_daxpy(n2,alpha,grad,incx,Xmat,incx);
		// Project
		for(i=0;i<n2;i++){
			if (Xmat[i]<Amat[i]-Bnd) Xmat[i]=Amat[i]-Bnd;
			if (Xmat[i]>Amat[i]+Bnd) Xmat[i]=Amat[i]+Bnd;}
		// Update stats and avergae	
		ovals[k]=sqrt(dabsf(eigv));
		alpha=k;cblas_dscal(n2,alpha,Xavmat,incx);
		alpha=1.0;cblas_daxpy(n2,alpha,Xmat,incx,Xavmat,incx);
		alpha=1.0/(k+1.0);cblas_dscal(n2,alpha,Xavmat,incx);
		// Report progress periodically 
		cputime=(double)((clock()-start_time)/CLOCKS_PER_SEC);
		cput[k]=cputime;
		if (k%Nperiod==0){
			// Compute gap and obj value.
			// Compute eig decomposition to form dual point
			nev=1;ncv=iminf(3,n); // TODO: Should get a better size estimate for ncv here (cf. arapck manual)
			arpinfo=simlrarpack(Xavmat,n,n,nev,ncv,tol,which,maxitrarp,0,&eigv,evec,&nconv,&nummatvec,&arcount);
			movals[k]=sqrt(dabsf(eigv));
			// Get sign of eigenvalues
			alpha=1.0;beta=0.0;
			cblas_dgemv(CblasColMajor,CblasNoTrans,n,n,alpha,Xavmat,n,evec,incx,beta,probv,incx);
			swsgn=dsignf(doubdot(probv,evec,n));
			if (eigv<=tol){swsgn=0.0;}
			// Form gradient (dual) matrix
			cblas_dcopy(n,evec,incx,bufmata,incx);
			alpha=swsgn;cblas_dscal(n,alpha,bufmata,incx);
			alpha=1.0;beta=0.0;
			cblas_dgemm(CblasColMajor,CblasNoTrans,CblasTrans,n,n,1,alpha,evec,n,bufmata,n,beta,grad,n);			
			gaps[k]=movals[k]; // Record gap
			gaps[k]-=doubdot(grad,Amat,n2);
			gaps[k]+=Bnd*doubasum(grad,n2);
			left_h=(int)floor(cputime/3600.0);left_m=(int)floor(cputime/60.0-left_h*60);left_s=(int)floor(cputime-left_h*3600-left_m*60);
			mexPrintf("Iter: %.3e   Obj: %.4e    Gap: %.4e   CPU Time: %2dh %2dm %2ds\n",(double)(k),movals[k],gaps[k],left_h,left_m,left_s);
			mexEvalString("drawnow;");
			//if (gaps[k]/gaps[0]<=targetg){break;}
			if (movals[k]<=targetg){break;}
		}}
	// Free everything
	free(evec);free(probv);free(bufmata);free(grad);
}