/* MEX function to solve SDP 

*/
// Last Modified: A. d'Aspremont, Feb 2009
// http://www.carva.org/alexandre.daspremont


#include "DetGrad.h" 

void mexFunction(int nlhs, mxArray *plhs[], int nrhs, const mxArray *prhs[])
{
	// For main loop
	double *Amat,*bvec,*cvec,*xvec,*avxvec,*grad,*ovals,*movals,*gaps,Bnd,stepscal,targetg,*cput;
	double *bufveca,*bufmat,alpha,stepsz,buf;
	int n,m,i,j,k,n2;
	int info,Nperiod,maxiter;
	double cputime;int left_h=0,left_m=0,left_s=0;
	clock_t start_time=(double)clock();
	// For LAPACK
	int work_size; 
	int lwork,inflapack,indmax;
	double *workvec;
	char jobz[1]="V",uplo[1]="U";
	// For Arpack & BLAS
	double *evec,eigv=0;
	int nummatvec,nconv,arcount,ncv,nev,incx=1,swsgn;
	char which[2]="LM";double tol=1e-8; int maxitrarp=1000;	
	
	// Inputs
	Amat=mxGetPr(prhs[0]);
	m = (int)mxGetN(prhs[0]);
	n2 = (int)mxGetM(prhs[0]);
	bvec = mxGetPr(prhs[1]);
	cvec = mxGetPr(prhs[2]);
	Bnd =(int)mxGetScalar(prhs[3]);
	stepscal =mxGetScalar(prhs[4]);
	maxiter =(int)mxGetScalar(prhs[5]);
	info =(int)mxGetScalar(prhs[6]);
	targetg =mxGetScalar(prhs[7]);
	
	n=(int)sqrt(n2);
	work_size=3*n+n*n;lwork=work_size;// For LAPACK
	// Outputs
	plhs[0] = mxCreateDoubleMatrix(m, 1, mxREAL);
	avxvec = mxGetPr(plhs[0]);
	plhs[1] = mxCreateDoubleMatrix(maxiter+1, 1, mxREAL);
	ovals = mxGetPr(plhs[1]);
	plhs[2] = mxCreateDoubleMatrix(maxiter+1, 1, mxREAL);
	movals = mxGetPr(plhs[2]);
	plhs[3] = mxCreateDoubleMatrix(maxiter+1, 1, mxREAL);
	gaps = mxGetPr(plhs[3]);
	plhs[4] = mxCreateDoubleMatrix(maxiter+1, 1, mxREAL);
	cput = mxGetPr(plhs[4]);



	// Allocate
	evec=(double*)calloc(n,sizeof(double));	
	bufveca=(double*)calloc(n,sizeof(double));
	xvec=(double*)calloc(m,sizeof(double));	
	grad=(double*)calloc(m,sizeof(double));	
	bufmat=(double*)calloc(n2,sizeof(double));	
	#ifndef usearpack 
	workvec=(double *) calloc(work_size,sizeof(double)); // For LAPACK
	#endif
	
	// Start main loop
	stepsz=stepscal/sqrt(maxiter);Nperiod=maxiter/info;
	alpha=0.0;cblas_dscal(m,alpha,xvec,incx);
	for(k=0;k<=maxiter;k++){
		// Form gradient
		alpha=0.0;cblas_dscal(n2,alpha,bufmat,incx);
		for(j=0;j<m;j++){alpha=xvec[j];cblas_daxpy(n2,alpha,Amat+j*n2,incx,bufmat,incx);}
		alpha=-1.0;cblas_daxpy(n2,alpha,cvec,incx,bufmat,incx);
		#ifdef usearpack // Switch between ARPACK and LAPACK
		nev=1;ncv=4; // TODO: Should get a better size estimate here (cf. arapck manual)
		info=simarpack(bufmat,n,nev,ncv,tol,which,maxitrarp,0,&eigv,evec,&nconv,&nummatvec,&arcount);
		#else
		dsyev(jobz,uplo,&n,bufmat,&n,bufveca,workvec,&lwork,&inflapack); 
		indmax=idxabsmax(bufveca,n);eigv=bufveca[indmax];cblas_dcopy(n,bufmat+indmax*n,incx,evec,incx);
		#endif
		for(i=0;i<n;i++){for(j=0;j<n;j++){bufmat[j*n+i]=evec[i]*evec[j];}}
		for(j=0;j<m;j++){grad[j]=doubdot(Amat+j*n2,bufmat,n2);}
		swsgn=dsignf(doubdot(xvec,grad,m)-doubdot(bufmat,cvec,n2));cblas_dscal(m,swsgn,grad,incx);
		alpha=-1.0;cblas_daxpy(m,alpha,bvec,incx,grad,incx);
		// Take step, project & record objective value + avg.
		alpha=-stepsz;cblas_daxpy(m,alpha,grad,incx,xvec,incx);
		buf=doubnorm2(xvec,m);if(buf>1e-12){alpha=dminif(Bnd,buf)/buf;cblas_dscal(m,alpha,xvec,incx);}
		ovals[k]=dabsf(eigv)-doubdot(bvec,xvec,m); 
		alpha=k;cblas_dscal(m,alpha,avxvec,incx);
		alpha=1.0;cblas_daxpy(m,alpha,xvec,incx,avxvec,incx);
		alpha=1.0/(k+1.0);cblas_dscal(m,alpha,avxvec,incx);		
		// Report progress periodically
		cputime=(double)((clock()-start_time)/CLOCKS_PER_SEC);
		cput[k]=cputime;
		if (k%Nperiod==0){
			alpha=0.0;cblas_dscal(n2,alpha,bufmat,incx);
			for(j=0;j<m;j++){alpha=avxvec[j];cblas_daxpy(n2,alpha,Amat+j*n2,incx,bufmat,incx);}
			alpha=-1.0;cblas_daxpy(n2,alpha,cvec,incx,bufmat,incx);
			#ifdef usearpack // Switch between ARPACK and LAPACK
			nev=1;ncv=4; // TODO: Should get a better size estimate here (cf. arapck manual)
			info=simarpack(bufmat,n,nev,ncv,tol,which,maxitrarp,0,&eigv,evec,&nconv,&nummatvec,&arcount);
			#else
			dsyev(jobz,uplo,&n,bufmat,&n,bufveca,workvec,&lwork,&inflapack); 
			indmax=idxabsmax(bufveca,n);eigv=bufveca[indmax];
			#endif
			movals[k]=dabsf(eigv)-doubdot(bvec,avxvec,m);
			// Compute gap
			for(i=0;i<n;i++){for(j=0;j<n;j++){bufmat[j*n+i]=evec[i]*evec[j];}}
			for(j=0;j<m;j++){grad[j]=doubdot(Amat+j*n2,bufmat,n2);}
			swsgn=dsignf(doubdot(xvec,grad,m)-doubdot(bufmat,cvec,n2));cblas_dscal(m,swsgn,grad,incx);
			alpha=-1.0;cblas_daxpy(m,alpha,bvec,incx,grad,incx);
			buf=0;for(j=0;j<n;j++){for(i=0;i<n;i++){buf+=cvec[j*n+i]*evec[i]*evec[j];}} // Could also recycle approx. above
			gaps[k]=movals[k]+swsgn*buf+Bnd*doubnorm2(grad,m);
			left_h=(int)floor(cputime/3600.0);left_m=(int)floor(cputime/60.0-left_h*60.0);left_s=(int)floor(cputime-left_h*3600-left_m*60);
			mexPrintf("Iter: %.3e   Obj: %.4e    Gap: %.4e   CPU Time: %2dh %2dm %2ds\n",(double)(k),movals[k],gaps[k],left_h,left_m,left_s);
			mexEvalString("drawnow;");
			//if (gaps[k]/gaps[0]<=targetg){break;}
			if (movals[k]<=targetg){break;}
		}}
	// Free everything
	free(evec);free(bufveca);free(bufmat);free(xvec);free(grad);
	#ifndef usearpack
	free(workvec);
	#endif
}

