#include "DetGrad.h"

// BLAS wrapper for win32 version
//#ifdef WIN32
void cblas_dscal( int N, double alpha, double *X, int incX)
{
	dscal(&N,&alpha,X,&incX);
}

void cblas_dcopy(int N,double *X,int incX,double *Y,int incY)
{
	dcopy(&N,X,&incX,Y,&incY);
}

void cblas_dgemm(enum CBLAS_ORDER Order,enum CBLAS_TRANSPOSE transA, enum CBLAS_TRANSPOSE transB,
                 int M, int N, int K, double alpha, double *A, int lda,
                 double *B, int ldb, double beta, double *C, int ldc)
{
	char ta[1],tb[1];
	if (transA==111)
	{
		*ta='N';
	}
	else
	{
		*ta='T';
	};
	if (transB==111) 
	{
		*tb='N';
	}
	else
	{
		*tb='T';
	};
	dgemm(ta,tb,&M,&N,&K,&alpha,A,&lda,B,&ldb,&beta,C,&ldc);
}

void cblas_dgemv(enum CBLAS_ORDER Order,enum CBLAS_TRANSPOSE transA,
                 int M, int N, double alpha, double *A, int lda,
                 double *B, int incB, double beta, double *C, int incC)
{
	char ta[1];
	if (transA==111)
	{
		*ta='N';
	}
	else
	{
		*ta='T';
	};
	dgemv(ta,&M,&N,&alpha,A,&lda,B,&incB,&beta,C,&incC);
}

void cblas_daxpy(int N,double alpha,double *X,int incX,double *Y,int incY)
{
	daxpy(&N,&alpha,X,&incX,Y,&incY);
}

void cblas_dger(enum CBLAS_ORDER Order,int m,int n,double alpha,double *x,int incx,double *y,int incy,double *A,int lda)
{
	dger(&m,&n,&alpha,x,&incx,y,&incy,A,&lda);
}
//#endif WIN32

// BLAS wrapper for the linux version
#ifdef linuxp
void cblas_dscal( int N, double alpha, double *X, int incX)
{
	dscal(&N,&alpha,X,&incX);
}

void cblas_dcopy(int N,double *X,int incX,double *Y,int incY)
{
	dcopy(&N,X,&incX,Y,&incY);
}

void cblas_dgemm(enum CBLAS_ORDER Order,enum CBLAS_TRANSPOSE transA, enum CBLAS_TRANSPOSE transB,
                 int M, int N, int K, double alpha, double *A, int lda,
                 double *B, int ldb, double beta, double *C, int ldc)
{
	char ta[1],tb[1];
	if (transA==111)
	{
		*ta='N';
	}
	else
	{
		*ta='T';
	};
	if (transB==111)
	{
		*tb='N';
	}
	else
	{
		*tb='T';
	};
	dgemm(ta,tb,&M,&N,&K,&alpha,A,&lda,B,&ldb,&beta,C,&ldc);
}

void cblas_dgemv(enum CBLAS_ORDER Order,enum CBLAS_TRANSPOSE transA,
                 int M, int N, double alpha, double *A, int lda,
                 double *B, int incB, double beta, double *C, int incC)
{
	char ta[1];
	if (transA==111)
	{
		*ta='N';
	}
	else
	{
		*ta='T';
	};
	dgemv(ta,&M,&N,&alpha,A,&lda,B,&incB,&beta,C,&incC);
}

void cblas_daxpy(int N,double alpha,double *X,int incX,double *Y,int incY)
{
	daxpy(&N,&alpha,X,&incX,Y,&incY);
}

void cblas_dger(enum CBLAS_ORDER Order,int m,int n,double alpha,double *x,int incx,double *y,int incy,double *A,int lda)
{
	dger(&m,&n,&alpha,x,&incx,y,&incy,A,&lda);
}
#endif

// Some useful functions ...
double doubsum(double *xmat, int n)
{
	int i;
	double res=0.0;
	for (i=0;i<n;i++){res+=xmat[i];};
	return res;
}

double doubdot(double *xvec, double *yvec, int n)
{
	int i;
	double res=0.0;
	for (i=0;i<n;i++){res+=xvec[i]*yvec[i];};
	return res;
}

int idxmax(double *xmat, int n)
{
	int i;
	int res=0;
	for (i=0;i<n;i++)
	{
		if (xmat[i]>xmat[res]) {res=i;}
	}
	return res;
}


int idxabsmax(double *xmat, int n)
{
	int i;
	int res=0;
	for (i=0;i<n;i++)
	{
		if (dabsf(xmat[i])>dabsf(xmat[res])) {res=i;}
	}
	return res;
}


double doubasum(double *xmat, int n)
{
	int i;
	double res=0.0;
	for (i=0;i<n;i++){res+=dabsf(xmat[i]);};
	return res;
}

double doubnorm2(double *xmat, int n)
{
	int i;
	double res=0.0;
	for (i=0;i<n;i++){res+=xmat[i]*xmat[i];};
	return sqrt(res);
}

double infnorm(double *xmat, int n)
{
	int i,j;
	double res=0.0,sum;
	
	for (j=0;j<n;j++){
		sum=0.0;
		for(i=0;i<n;i++)
			sum+=dabsf(xmat[j+i*n]);
		if(sum>=res) res=sum;
	}	
	return res;
}

double frobnorm(double *xmat, int n)
{
	int i,j;
	double res=0.0;
	
	for (i=0;i<n;i++)
		for (j=0;j<n;j++)
			res+=(xmat[i*n+j]*xmat[i*n+j]);
	
	return pow(res,.5);
}

double dsignf(double x)
{
	if (x>=0)
		return 1.0;
	else
		return -1.0;
}

double dminif(double x, double y)
{
	if (x>=y)
		return y;
	else
		return x;
}

double dmaxf(double x, double y)
{
	if (x>=y)
		return x;
	else
		return y;
}

int imaxf(int x, int y)
{
	if (x>=y)
		return x;
	else
		return y;
}

double dabsf(double x)
{
	if (x>=0)
		return x;
	else
		return -x;
}

void dispmat(double *xmat, int n, int m)
{
	int i,j;
	
	for (i=0; i<n; i++)
	{
		for (j=0;j<m;j++)
		{
			printf("%+.4f ",xmat[j*n+i]);
		}
		printf("\n");
	}
	printf("\n");
}

// Simple wrapper for calls to ARPACK 
int simarpack(double *Amat, int n, int nev, int ncv, double tol, char* which, int maxitr, int printlevel, double* evals, double* evecs, int* nconv, int* nummatvec, int *count)
{
	// From ARPACK examples. See dsvrv1.f, ... 
	double *v, *workl, *workd, *d, *resid;
	int *select, iparam[11], ipntr[11];
	char bmat[2]="I",all[4]="All"; // ARPACK params we only solve simple eigenvalue problems
	int ldv, ido, lworkl, info, ierr;
	int mode, ishfts, rvec,i,incx=1;
	double sigma,zero,minevalest;
	double alpha=1.0,beta=0.0;

	// Internal ARPACK parameters and vars
	ldv=n;zero=0.0;lworkl=ncv*(ncv+10);
	info=0;ido=0;ishfts =1;mode=1;rvec=1;
	iparam[1-1]=ishfts;iparam[3-1]=maxitr;iparam[7-1]=mode;
	*nummatvec=0;
	
	// Allocate memory for arpack workspace
	v = (double*)calloc(ldv*ncv, sizeof(double)); 
	workl = (double*)calloc(lworkl, sizeof(double));
	workd = (double*)calloc(3*n, sizeof(double));
	d = (double*)calloc(ncv*2, sizeof(double));
	resid = (double*)calloc(n, sizeof(double));
	select = (int*)calloc(ncv, sizeof(int));
		
	// ARPACK loop
	*count=0;
	do {
		(*count)++;
		dsaupd(&ido, bmat, &n, which, &nev, &tol, resid, &ncv, v, &ldv, iparam, ipntr, workd, workl, &lworkl, &info);
		//if (iparam[3-1]>1000) printf("ARPACK dsaupd number iters: %d.\n", iparam[3-1]);
		if(ido == -1 || ido == 1) {
			// *** Matrix vector product, X is at workd+ipntr[1-1]-1 and the product Y is at workd+ipntr[2-1]-1 ***
			cblas_dgemv(CblasColMajor,CblasNoTrans,n,n,alpha,Amat,n,workd + ipntr[1-1] - 1,incx,beta,workd + ipntr[2-1] - 1,incx);
			*nummatvec = *nummatvec + 1;}
		else if(info != 0) {
			if(printlevel > 0) printf("ARPACK dsaupd error (info = %d).\n", info);}} 
	while(ido != 99);
	// Post processing
	dseupd(&rvec, all, select, d, v, &ldv, &sigma, bmat, &n, which, &nev, &tol, resid, &ncv, v, &ldv, iparam, ipntr, workd, workl, &lworkl, &ierr);
	if(ierr != 0) {
		if(printlevel > 0) printf("ARPACK dseupd error (ierr = %d)\n",ierr);
		info = ierr;}
	*nconv = iparam[5-1];
	if(printlevel > 0 && *nconv < nev) {
		printf("Warning: %d out of %d evals converged. dsaupd called %d times\n", *nconv, nev,*count);
		if(nev == 1) printf("Best estimate returned.\n");
		else printf("\n");}
	else{// Store eigensystem
		cblas_dcopy(nev,d,1,evals,1);
		cblas_dcopy(nev*n,v,1,evecs,1);
		if(nev == 1 && *nconv == 0) {
			minevalest = 1.0e10;
			for(i = 1; i <= ncv; i++) minevalest = -imaxf(-minevalest,-workl[ipntr[6-1]-2+i]);
			evals[0] = minevalest;}}

	// Free everything
	free(select);
	free(resid);
	free(d);
	free(workd);
	free(workl);
	free(v);
	// Return status
	if(*nconv < nev) return -1; // Not all eigenvalues were found
	return info;  // Return arapck info otherwise
}

// Simple wrapper for calls to ARPACK on sparse matrices
int simsparpack(double *sr, int *irc, int *jcs, int n, int nev, int ncv, double tol, char* which, int maxitr, int printlevel, double* evals, double* evecs, int* nconv, int* nummatvec, int *count)
{
	// From ARPACK examples. See dsvrv1.f, ... 
	double *v, *workl, *workd, *d, *resid;
	double *xv,*yv;
	int *select, iparam[11], ipntr[11];
	char bmat[2]="I",all[4]="All"; // ARPACK params we only solve simple eigenvalue problems
	int ldv, ido, lworkl, info, ierr;
	int mode, ishfts, rvec,i,k,incx=1;
	double rhs;
	double sigma,zero,minevalest;
	double alpha=1.0;
	
	// Internal ARPACK parameters and vars
	ldv=n;zero=0.0;lworkl=ncv*(ncv+10);
	info=0;ido=0;ishfts =1;mode=1;rvec=1;
	iparam[1-1]=ishfts;iparam[3-1]=maxitr;iparam[7-1]=mode;
	*nummatvec=0;
	
	// Allocate memory for arpack workspace
	v = (double*)calloc(ldv*ncv, sizeof(double)); 
	workl = (double*)calloc(lworkl, sizeof(double));
	workd = (double*)calloc(3*n, sizeof(double));
	d = (double*)calloc(ncv*2, sizeof(double));
	resid = (double*)calloc(n, sizeof(double));
	select = (int*)calloc(ncv, sizeof(int));
	
	// ARPACK loop
	*count=0;
	do {
		(*count)++;
		dsaupd(&ido, bmat, &n, which, &nev, &tol, resid, &ncv, v, &ldv, iparam, ipntr, workd, workl, &lworkl, &info);
		//if (iparam[3-1]>1000) printf("ARPACK dsaupd number iters: %d.\n", iparam[3-1]);
		if(ido == -1 || ido == 1) {
			// *** Matrix vector product, X is at workd+ipntr[1-1]-1 and the product Y is at workd+ipntr[2-1]-1 ***
			xv=workd+ipntr[1-1]-1;yv=workd+ipntr[2-1]-1;
			alpha=0.0;cblas_dscal(n,alpha,yv,incx);
			//cblas_dgemv(CblasColMajor,CblasNoTrans,n,n,alpha,Amat,n,workd + ipntr[1-1] - 1,incx,beta,workd + ipntr[2-1] - 1,incx);
			for (i=0;i<n;i++){ /* Loop through columns */
				rhs=xv[i];
				for (k=jcs[i];k<jcs[i+1];k++) {yv[irc[k]]+=sr[k]*rhs;} /* Loop through non-zeros in ith column */		
			}			
		*nummatvec = *nummatvec + 1;}
		else if(info != 0) {
		if(printlevel > 0) printf("ARPACK dsaupd error (info = %d).\n", info);}} 
	while(ido != 99);
	// Post processing
	dseupd(&rvec, all, select, d, v, &ldv, &sigma, bmat, &n, which, &nev, &tol, resid, &ncv, v, &ldv, iparam, ipntr, workd, workl, &lworkl, &ierr);
	if(ierr != 0) {
		if(printlevel > 0) printf("ARPACK dseupd error (ierr = %d)\n",ierr);
	info = ierr;}
	*nconv = iparam[5-1];
	if(printlevel > 0 && *nconv < nev) {
		printf("Warning: %d out of %d evals converged. dsaupd called %d times\n", *nconv, nev,*count);
		if(nev == 1) printf("Best estimate returned.\n");
	else printf("\n");}
	else{// Store eigensystem
		cblas_dcopy(nev,d,1,evals,1);
		cblas_dcopy(nev*n,v,1,evecs,1);
		if(nev == 1 && *nconv == 0) {
			minevalest = 1.0e10;
			for(i = 1; i <= ncv; i++) minevalest = -imaxf(-minevalest,-workl[ipntr[6-1]-2+i]);
		evals[0] = minevalest;}}
	
	// Free everything
	free(select);
	free(resid);
	free(d);
	free(workd);
	free(workl);
	free(v);
	// Return status
	if(*nconv < nev) return -1; // Not all eigenvalues were found
	return info;  // Return arapck info otherwise
}
