/* MEX function to solve SDP 

*/
// Last Modified: A. d'Aspremont, Feb 2009
// http://www.carva.org/alexandre.daspremont


#include "DetTraceNorm.h" 

void mexFunction(int nlhs, mxArray *plhs[], int nrhs, const mxArray *prhs[])
{
	// For main loop
	double *sr;int *irs,*jcs;
	double *Xmat,*Xavmat,*grad,*ovals,*movals,*gaps,Bnd,stepscal,*cput,targetg;
	double *probv,*bufmata;
	double alpha,beta,stepsz;
	int n,i,j,k,l,ks,n2,rnk;
	int info,arpinfo,Nperiod,maxiter;
	double cputime;int left_h=0,left_m=0,left_s=0;
	clock_t start_time=(double)clock();
	double *evec,*eigv,*swsgn;
	int nummatvec,nconv,arcount,ncv,nev,incx=1;
	char which[2]="LM";double tol=1e-6; int maxitrarp=1000; // ARPACK parameters
		
	// Inputs
	sr  = mxGetPr(prhs[0]); // Get sparse ratings matrix
	irs = mxGetIr(prhs[0]);
	jcs = mxGetJc(prhs[0]);
	n = (int)mxGetN(prhs[0]);
	rnk =(int)mxGetScalar(prhs[1]); // Upper bound on target rank
	Bnd =(int)mxGetScalar(prhs[2]); // Bound on ratings
	stepscal = mxGetScalar(prhs[3]);
	maxiter =(int)mxGetScalar(prhs[4]);
	info =(int)mxGetScalar(prhs[5]);
	targetg =mxGetScalar(prhs[6]);
	n2=n*n;
	// Outputs
	plhs[0] = mxCreateDoubleMatrix(n,n, mxREAL);
	Xmat = mxGetPr(plhs[0]);
	plhs[1] = mxCreateDoubleMatrix(n,n, mxREAL);
	Xavmat = mxGetPr(plhs[1]);
	plhs[2] = mxCreateDoubleMatrix(maxiter+1, 1, mxREAL);
	ovals = mxGetPr(plhs[2]);
	plhs[3] = mxCreateDoubleMatrix(maxiter+1, 1, mxREAL);
	movals = mxGetPr(plhs[3]);
	plhs[4] = mxCreateDoubleMatrix(maxiter+1, 1, mxREAL);
	gaps = mxGetPr(plhs[4]);
	plhs[5] = mxCreateDoubleMatrix(maxiter+1, 1, mxREAL);
	cput = mxGetPr(plhs[5]);


	// Allocate
	evec=(double*)calloc(n*rnk,sizeof(double));
	probv=(double*)calloc(n,sizeof(double));
	bufmata=(double*)calloc(n*rnk,sizeof(double));
	eigv=(double*)calloc(rnk,sizeof(double));
	swsgn=(double*)calloc(rnk,sizeof(double));
	grad=(double*)calloc(n*n,sizeof(double));	

	// Initialize X matrix
	for(i=0;i<n2;i++){Xmat[i]=1.0;}
	for (j=0;j<n;j++){ /* Loop through columns */
		for (ks=jcs[j];ks<jcs[j+1];ks++) {Xmat[j*n+irs[ks]]=sr[ks];}} /* Loop through non-zeros in jth column */		
	// Start main loop
	stepsz=stepscal/sqrt(maxiter);Nperiod=maxiter/info;
	for(k=0;k<=maxiter;k++){
		// Compute eig decomposition
		nev=rnk;ncv=iminf(3*rnk,n); // TODO: Should get a better size estimate for ncv here (cf. arapck manual)
		arpinfo=simlrarpack(Xmat,n,n,nev,ncv,tol,which,maxitrarp,0,eigv,evec,&nconv,&nummatvec,&arcount);
		// Get gradient and step
		// Get sign of eigenvalues
		for(l=0;l<rnk;l++){
			alpha=1.0;beta=0.0;
			cblas_dgemv(CblasColMajor,CblasNoTrans,n,n,alpha,Xmat,n,evec+l*n,incx,beta,probv,incx);
			swsgn[l]=dsignf(doubdot(probv,evec+l*n,n));
			if (eigv[l]<=tol){swsgn[l]=0.0;}}
		// Form gradient matrix
		cblas_dcopy(n*rnk,evec,incx,bufmata,incx);
		for(l=0;l<rnk;l++){alpha=swsgn[l];cblas_dscal(n,alpha,bufmata+l*n,incx);}
		alpha=1.0;beta=0.0;
		cblas_dgemm(CblasColMajor,CblasNoTrans,CblasTrans,n,n,rnk,alpha,evec,n,bufmata,n,beta,grad,n);
		// Take step 
		alpha=-stepsz;cblas_daxpy(n2,alpha,grad,incx,Xmat,incx);
		// Project on ratings
		for (j=0;j<n;j++){ /* Loop through columns */
			for (ks=jcs[j];ks<jcs[j+1];ks++) {Xmat[j*n+irs[ks]]=sr[ks];}} /* Loop through non-zeros in jth column */
		for(i=0;i<n2;i++){if (Xmat[i]<0.0) Xmat[i]=0.0;}
		for(i=0;i<n2;i++){if (Xmat[i]>Bnd) Xmat[i]=Bnd;}
		// Update stats and avergae	
		ovals[k]=0;
		for(l=0;l<rnk;l++){ovals[k]+=sqrt(dabsf(eigv[l]));}
		alpha=k;cblas_dscal(n2,alpha,Xavmat,incx);
		alpha=1.0;cblas_daxpy(n2,alpha,Xmat,incx,Xavmat,incx);
		alpha=1.0/(k+1.0);cblas_dscal(n2,alpha,Xavmat,incx);
		// Report progress periodically 
		cputime=(double)((clock()-start_time)/CLOCKS_PER_SEC);
		cput[k]=cputime;
		if (k%Nperiod==0){
			// Compute gap and obj. value
			// Compute eig decomposition to form dual point
			nev=rnk;ncv=iminf(3*rnk,n); // TODO: Should get a better size estimate for ncv here (cf. arapck manual)
			arpinfo=simlrarpack(Xavmat,n,n,nev,ncv,tol,which,maxitrarp,0,eigv,evec,&nconv,&nummatvec,&arcount);
			for(l=0;l<rnk;l++){movals[k]+=sqrt(dabsf(eigv[l]));}
			// Get sign of eigenvalues
			for(l=0;l<rnk;l++){
				alpha=1.0;beta=0.0;
				cblas_dgemv(CblasColMajor,CblasNoTrans,n,n,alpha,Xavmat,n,evec+l*n,incx,beta,probv,incx);
				swsgn[l]=dsignf(doubdot(probv,evec+l*n,n));
				if (eigv[l]<=tol){swsgn[l]=0.0;}}
			// Form gradient (dual) matrix
			cblas_dcopy(n*rnk,evec,incx,bufmata,incx);
			for(l=0;l<rnk;l++){alpha=swsgn[l];cblas_dscal(n,alpha,bufmata+l*n,incx);}
			alpha=1.0;beta=0.0;
			cblas_dgemm(CblasColMajor,CblasNoTrans,CblasTrans,n,n,rnk,alpha,evec,n,bufmata,n,beta,grad,n);			
			gaps[k]=movals[k]; // Record gap
			for(j=0;j<n;j++){for(ks=jcs[j];ks<jcs[j+1];ks++){gaps[k]-=grad[j*n+irs[ks]]*sr[ks];grad[j*n+irs[ks]]=0.0;}}
			for(j=0;j<n2;j++){gaps[k]-=Bnd*dminif(0.0,grad[j]);}			
			left_h=(int)floor(cputime/3600.0);left_m=(int)floor(cputime/60.0-left_h*60);left_s=(int)floor(cputime-left_h*3600-left_m*60);
			mexPrintf("Iter: %.3e   Obj: %.4e    Gap: %.4e   CPU Time: %2dh %2dm %2ds\n",(double)(k),movals[k],gaps[k],left_h,left_m,left_s);
			mexEvalString("drawnow;");
			//if (gaps[k]/gaps[0]<=targetg){break;}
			if (movals[k]<=targetg){break;}
		}}
	// Free everything
	free(evec);free(probv);free(bufmata);free(eigv);free(swsgn);free(grad);
}